//========================================================= file = run.c ======
//=  Run function for REMOTE++                                                =
//=============================================================================
//=  Notes: 1) Input files must be different for all jobs in joblist          =
//=         2) Assumes c:/temp exists on remote machines                      =
//=         3) Assumes c:/Remotepp exists and it contains remotepp.exe and    =
//=             files to be copied to the remote machines                     =
//=         4) Hostlist.txt, Joblist.txt, and Status.txt must exist           =
//=         5) If only 1 host and it is invalid will get stuck                =
//=         6) An unused host is always "valid" in the status file            =
//=---------------------------------------------------------------------------=
//=  Contains: run() - Runs programs in joblist on remote hosts in hostlist   =
//=---------------------------------------------------------------------------=
//=  History:  AMH (11/01/02) - Genesis (from mrun.c by KJC)                  =
//=============================================================================
#define EXTERNAL

//----Include files------------------------------------------------------------
#include "remotepp.h"

//=============================================================================
//= Function to run programs on remote hosts                                  =
//=============================================================================
int run(void)
{
  FILE          *st1;                      // File stream #1 for JOB_LIST
  FILE          *st2;                      // File stream #2 for HOST_LIST
  FILE          *st3;                      // File stream #3 for STATUS
  FILE          *st4;                      // File stream #4 for exe_file
  FILE          *st5;                      // File stream #5 for in_file
  FILE          *st6;                      // File stream #6 for out_file
  char          run_mode[10];              // Mode of the input & output
  char          host_name[100];            // Remote host name
  char          exe_file[100];             // Executable file name
  char          in_file[100];              // Input file name
  char          out_file[100];             // Output file name
  int           i, j;                      // Loop counters
  int           done_flag;                 // Done flag for job list
  int           assign_flag;               // Assignment flag
  int           job_number;                // number of job being processed 
  HANDLE        handle_array[MAX_JOBS];    // array of handles
  int		file_error;		   // Error in opening file

  // Output banner
  get_time(Time);
  printf("Starting master for remote runs at %s... \n", Time);
  printf("  >>> Ctrl-C to break-out <<< \n");

  // Open job_list
  st1 = fopen(JOB_LIST, "r");
  if (st1 == NULL)
  {
    printf("*** ERROR in run() - unable to open job list %s \n", JOB_LIST);
    return(-1);
  }

  // Open host_list
  st2 = fopen(HOST_LIST, "r");
  if (st2 == NULL)
  {
    printf("*** ERROR in run() - unable to open host list %s \n", HOST_LIST);
    return(-1);
  }

  //Open the STATUS file
  st3 = fopen(STATUS, "w");
  if (st3 ==NULL)
  {
    printf("*** ERROR in run() - unable to open status file %s \n", STATUS);
    return(-1);
  }

  // Build the run list
  Num_jobs = 0;
  Num_hosts = 0;

  while(TRUE)
  {
    //initially set error to 0
    file_error=0;

    //scan in job to be executed
    fscanf(st1, "%s %s %s %s", run_mode, exe_file, in_file, out_file);
    if (feof(st1))
      break;
   
    //Set the Mode
    if((strcmp(run_mode, "file")==0)||(strcmp(run_mode,"File")==0)
	||(strcmp(run_mode, "FILE")==0))
    {  
      Run_list[Num_jobs].mode = CLASSIC;
      fprintf(st3, "Mode is classic. \n");
    }

    else if ((strcmp(run_mode, "std")==0)||(strcmp(run_mode,"Std")==0)
	||(strcmp(run_mode, "STD")==0))
    {  
      Run_list[Num_jobs].mode = NEW;
      fprintf(st3, "Mode is new. \n");
    }
    
    else
    {
      fprintf(st3, "Mode %s is not valid. \n", run_mode);
      file_error=1;
    }

    // Determine if the executable exists
    st4 = fopen(exe_file, "r");
    if (st4 == NULL)
    {
      file_error=1;
      fprintf(st3,"Executable file %s could not be found. \n", exe_file);
    }
    else fprintf(st3, "Executable file %s found. \n", exe_file);
    fclose(st4);

    // Determine if input file exists
    st5 = fopen(in_file, "r");
    if (st5 == NULL)
    {
      file_error=1;      
      fprintf(st3, "Input file %s could not be found. \n", in_file);
    }
    else fprintf(st3, "Input file %s found. \n", in_file);
    fclose(st5);

    // Determine if output file exists
    st6 = fopen(out_file, "r");
    if (st6 == NULL)
    {
      file_error=1;
      fprintf(st3, "Output file %s could not be found. \n", out_file);
    }
    else fprintf(st3, "Output file %s found. \n", out_file);

    //print if job was run
    if (file_error==0)
      fprintf(st3, "Job was run\n");
    else
      fprintf(st3, "Job was not run\n");

    fprintf(st3, "\n");

    // Print dummy message to output file
    fprintf(st6, "Dummy");
    fclose(st6);


    //Set up runlist if all three files exist
    if(file_error==0)
    { 
      strcpy(Run_list[Num_jobs].exe_file, exe_file);
      strcpy(Run_list[Num_jobs].in_file, in_file);
      strcpy(Run_list[Num_jobs].out_file, out_file);

      fscanf(st2, "%s", host_name);
      if (!feof(st2))
      {
        strcpy(Hosts[Num_hosts].host_name, host_name);
        Num_hosts++;
      }
      
      if (feof(st2))
        strcpy(host_name, "");
      strcpy(Run_list[Num_jobs].host_name, host_name);
      

      if (strcmp(host_name, "") != 0)
      {
        Run_list[Num_jobs].status = ASSIGNED;
        Hosts[Num_hosts-1].status = BUSY;
      }
      else
        Run_list[Num_jobs].status = UNASSIGNED;

      Num_jobs++;
    }
  }

  //Copy the remainder of the hosts into the Hosts array
  fscanf(st2, "%s", host_name);
  while(!feof(st2))
  {
    strcpy(Hosts[Num_hosts].host_name, host_name);
    Hosts[Num_hosts].status = AVAILABLE;
    Num_hosts++;
    fscanf(st2, "%s", host_name);
  }

  // Close JOB_LIST and HOST_LIST
  fclose(st1);
  fclose(st2);

  // Run the jobs
  done_flag = FALSE;
  while(done_flag == FALSE)
  {
    // Test if an assignment can be made
    assign_flag = FALSE;
    for (i=0; i<Num_jobs; i++)
    {
      if (((strcmp(Run_list[i].host_name, "") != 0) &&
      (Run_list[i].status == ASSIGNED)))
      {
        job_number=i;
        Run_list[i].status = RUNNING;
        assign_flag = TRUE;
      }
    
      // Make an assignment if it can be done
      if (assign_flag == TRUE)
      {
        // Output a job assigned banner
        get_time(Time);
        printf("Running (%s %s %s) on %s at %s \n", Run_list[job_number].exe_file,
          Run_list[job_number].in_file, Run_list[job_number].out_file,
          Run_list[job_number].host_name, Time);

        // Lanch the thread for the job
        handle_array[job_number] = (HANDLE)_beginthread(run_thread, 		4096,(void*)job_number);
      }
      assign_flag=FALSE;
    }
    // Test if all jobs done
    done_flag = TRUE;
    for (i=0; i<Num_jobs; i++)
    {
      if (Run_list[i].status != DONE)
      {
        done_flag = FALSE;

        //Assign a host if needed and available
        if(strcmp(Run_list[i].host_name, "")==0)
        {
          for(j=0; j<Num_hosts; j++)
            if(Hosts[j].status==AVAILABLE)
            {
              strcpy(Run_list[i].host_name, Hosts[j].host_name);
              Hosts[j].status = BUSY;
              Run_list[i].status = ASSIGNED;
              break;
            }
        }
      }
    }
  }

  //Print host status to status.txt
  for(i=0; i<Num_hosts; i++)
  {
    if(Hosts[i].status == INVALID)
      fprintf(st3,"%s is not a valid host.\n", Hosts[i].host_name);
    else
      fprintf(st3,"%s is a valid host.\n", Hosts[i].host_name);
  }
  fprintf(st3, "\nThere are %d valid jobs.\n", Num_jobs);
  fclose(st3);

  return(1);
}

//===========================================================================
//=  Thread code for run()                                                  =
//===========================================================================
void run_thread(void *thread_arg)
{
  int          current_job;                // number of the current job 
  char         rcp_exe[100];               // rcp command to copy exefile
  char         rcp_in[100];                // rcp command to copy infile
  char         rcp_out[100];               // rcp command to copy outfile
  char         rshstring[100];             // rsh command to execute exefile
  int          i, j;                       // Loop counters
  int          retcode;                    // Return code

  //set thread_arg to struct1
  current_job= (int) thread_arg;

  //form rcp string to copy files to remote host
  strcpy(rcp_exe, "rcp -b /Remotepp/");
  strcat(rcp_exe, Run_list[current_job].exe_file);
  strcat(rcp_exe, " ");
  strcat(rcp_exe, Run_list[current_job].host_name);
  strcat(rcp_exe, ":/temp");

  // Copy the executable file to the remote host
  retcode = system(rcp_exe);

  // Check if host in hostlist is valid
  if (retcode > 0)
  {
    for(i=0; i<Num_hosts; i++)
    {
      if(strcmp(Run_list[current_job].host_name, Hosts[i].host_name)==0)
      {
        Hosts[i].status = INVALID;
        break;
      }
    }
    strcpy(Run_list[current_job].host_name, "");
    Run_list[current_job].status = UNASSIGNED;
    _endthread();
  }

  // Set up and send job if mode is New
  if(Run_list[current_job].mode == NEW)
  {
    //form rsh string to execute on remote host
    strcpy(rshstring, "rsh ");
    strcat(rshstring, Run_list[current_job].host_name);
    strcat(rshstring, " c:/temp/");
    strcat(rshstring, Run_list[current_job].exe_file);
    strcat(rshstring, " <");
    strcat(rshstring, Run_list[current_job].in_file);
    strcat(rshstring, " >");
    strcat(rshstring, Run_list[current_job].out_file);
    
    // run rsh with redirection of input and output
    system(rshstring);
  }
  
  // Set up and send job if mode is Classic
  else
  {
    //form rcp string to copy input file to remote host
    strcpy(rcp_in, "rcp -b /Remotepp/");
    strcat(rcp_in, Run_list[current_job].in_file);
    strcat(rcp_in, " ");
    strcat(rcp_in, Run_list[current_job].host_name);
    strcat(rcp_in, ":/temp/infile");

    //rcp the input file
    system(rcp_in);

    //form rcp string to copy output file to remote host
    strcpy(rcp_out, "rcp -b /Remotepp/");
    strcat(rcp_out, Run_list[current_job].out_file);
    strcat(rcp_out, " ");
    strcat(rcp_out, Run_list[current_job].host_name);
    strcat(rcp_out, ":/temp/outfile");

    //rcp the output file to remote
    system(rcp_out);

    //form the rsh string
    strcpy(rshstring, "rsh ");
    strcat(rshstring, Run_list[current_job].host_name);
    strcat(rshstring, " c:/temp/");
    strcat(rshstring, Run_list[current_job].exe_file);

    // run rsh with input and output from file
    system(rshstring);

    //form rcp string to copy output file back from remote host
    strcpy(rcp_out, "rcp -b ");
    strcat(rcp_out, Run_list[current_job].host_name);
    strcat(rcp_out, ":/temp/outfile");
    strcat(rcp_out, " ");
    strcat(rcp_out, "/Remotepp/");
    strcat(rcp_out, Run_list[current_job].out_file);

    //rcp the output file back from remote
    system(rcp_out);
  }
 
  //Output a done message
  get_time(Time);
  printf("Job with outfile %s completed at %s \n",
    Run_list[current_job].out_file, Time);

  // Set status to DONE
  Run_list[current_job].status = DONE;

  //Set host available
  for(j=0; j<Num_hosts; j++)
  {
    if(strcmp(Run_list[current_job].host_name, Hosts[j].host_name)==0)
    {
     Hosts[j].status=AVAILABLE;
     break;
    }
  }
  _endthread();
}
